#include "tricks.h"

#include <conio.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#define PI 3.1415927

void load( char *file, byte *data, unsigned int length );
void loadpal( char *file );

void init();
void rotate();
void project();

byte buffer[64000];
byte Texture[4096];
byte Txt[4096];                 // roterad textur..

byte pal[255][3];
int sintab[1280];              // tabell med sin(x)<<16.. 0-1024

struct
{
   int x,y;
   float fx,fy,fz;
} pnt[8];

struct
{
   float fx,fy,fz;
} kubp[8];

int plane[6][4];

struct
{
   int beg, end;

   int btx, bty;
   int etx, ety;

} scanline[200];

// **********

void edge( int p1, int p2, int e )
{
   int tx1, ty1, tx2, ty2;

   switch(e)
   {
      case 0:  tx1 = ty1 = tx2 = 0;
               ty2 = 63;
               break;

      case 1:  ty1 = tx2 = ty2 = 63;
               tx1 = 0;
               break;

      case 2:  tx1 = ty1 = tx2 = 63;
               ty2 = 0;
               break;

      case 3:  ty1 = tx2 = ty2 = 0;
               tx1 = 63;
               break;
   }

   int y1 = pnt[p1].y;

   if( y1 == pnt[p2].y )
   {
      scanline[ y1 ].beg = pnt[p1].x;
      scanline[ y1 ].btx = tx1;
      scanline[ y1 ].bty = ty1;

      scanline[ y1 ].end = pnt[p2].x;
      scanline[ y1 ].etx = tx2;
      scanline[ y1 ].ety = ty2;
   }
   else
   {
      int y2 = pnt[p2].y;

      int x1 = pnt[p1].x;
      int x2 = pnt[p2].x;

      if( y2 < y1 )             // r det bakvnt?
      {
         swap( tx1, tx2 );
         swap( ty1, ty2 );
         swap( x1 , x2);
         swap( y1 , y2);
      }

      int dy = y2-y1;           // skillnad i y-led

   // *** steps ***

      int stx = ((tx2 - tx1) << 8) / dy;
      int sty = ((ty2 - ty1) << 8) / dy; 

      int step = ((x2 - x1) << 8) / dy;

   // *** values ***

      tx1 <<= 8;
      ty1 <<= 8;
      x1  <<= 8;
      x1   += step;

      for( int Y = (y1+1); Y <= y2; Y++)
      {
         scanline[Y].end = (x1 >> 8);

         scanline[Y].etx = (tx1 >> 8);
         scanline[Y].ety = (ty1 >> 8);

         if( scanline[Y].beg == -1 )
         {
            scanline[Y].beg = (x1 >> 8);

            scanline[Y].btx = (tx1 >> 8);
            scanline[Y].bty = (ty1 >> 8);
         }

         x1 += step;
         tx1 += stx;
         ty1 += sty;
      }         
   }
}   

void draw( int f )
{
   for( int Y=0; Y<200; Y++)
      scanline[Y].beg = scanline[Y].end = -1;      

   edge( plane[f][0], plane[f][1], 0);
   edge( plane[f][1], plane[f][2], 1);
   edge( plane[f][2], plane[f][3], 2);
   edge( plane[f][3], plane[f][0], 3);

   for( Y=0; Y<200; Y++)
   {
      int x1 = scanline[Y].beg;

      if( x1 != -1 )
      {
         int dx  = scanline[Y].end - x1;
         int etx = scanline[Y].etx << 8;
         int ety = scanline[Y].ety << 6;
         int TX  = scanline[Y].btx << 8;
         int TY  = scanline[Y].bty << 6;

         if( dx == 0 )
            dx = 1;
         else if (dx < 0)
         {
            x1 = scanline[Y].end;
            swap( etx, TX );
            swap( ety, TY );
            dx = -dx;
         }   

         byte *buf = buffer + (Y<<8) + (Y<<6) + x1;

         int sx = (etx - TX) / dx;
         int sy = (ety - TY) / dx;

         for( int X = 0; X <= dx; X++)
         {
            *buf++ = Txt[ (TY & 0xffffffc0) + (TX>>8) ];
            TX += sx;
            TY += sy;
         }
      }
   }
}

int main()
{
   init();
   do
   {
      rotate();
      project();
      Pal(0,0,20,0);

      for( int i=0; i<6; i++ )
      {
         int p0 = plane[i][0];
         int p1 = plane[i][1];
         int p2 = plane[i][2];
         int p3 = plane[i][3];
            
         float norm =   ((pnt[p3].fy - pnt[p0].fy) * (pnt[p1].fx - pnt[p0].fx))
                      - ((pnt[p1].fy - pnt[p0].fy) * (pnt[p3].fx - pnt[p0].fx));

         if( norm< -0.22 )
            draw( i );
      }

      Pal(0,0,0,0);
      WaitRetrace();

      int *v= (int*)vga;
      int *b= (int*)buffer;
   
      Pal(0,20,0,0);
      for( i=0; i<16000; i++)
      {
         *(v++) = *b;
         *(b++) = 0;
      }

      Pal(0,0,0,20);

   } while ( !kbhit() );

   SetText();

   return 0;
}

void initpoint( int p, float x, float y, float z)
{
   kubp[p].fx = x;
   kubp[p].fy = y;
   kubp[p].fz = z;
}

void initface ( int p, int a, int b, int c, int d)
{
   plane[p][0] = a;
   plane[p][1] = b;
   plane[p][2] = c;
   plane[p][3] = d;
}
   

void init()
{
   load( "fz-text.raw", Texture, 4096 );
   loadpal( "frozty.pal" );

   SetMCGA();

   for( int i=0; i<256; i++ )
      Pal(i, pal[i][0], pal[i][1], pal[i][2] );

   initpoint( 0,  1,  1, 1 );
   initpoint( 1,  1, -1, 1 );
   initpoint( 2, -1, -1, 1 );
   initpoint( 3, -1,  1, 1 );

   initpoint( 4,  1,  1, -1 );
   initpoint( 5,  1, -1, -1 );
   initpoint( 6, -1, -1, -1 );
   initpoint( 7, -1,  1, -1 );

   initface( 0, 0,3,2,1);
   initface( 1, 4,0,1,5);
   initface( 2, 7,4,5,6);
   initface( 3, 3,7,6,2);
   initface( 4, 7,3,0,4);
   initface( 5, 5,1,2,6);

   for( i=0; i<1280; i++)      
      sintab[i] = (int) (cos( ((float) i) * 2 * PI /1024 )*65536);

}


void rotate()
{
   static float Xangle = 0;
   static float Yangle = 0;
   static float Zangle = 0;

   Xangle += 0.01;
   Yangle += 0.02;
   Zangle += 0.03;

   float sinX = sin(Xangle);
   float cosX = cos(Xangle);

   float sinY = sin(Yangle);
   float cosY = cos(Yangle);

   float sinZ = sin(Zangle);
   float cosZ = cos(Zangle);

   static double zoom = 0;

   zoom += .04;
   float z = sin(zoom) / 3 + .9;

   for( int i=0; i<8; i++)
	{
      pnt[i].fy = kubp[i].fy * cosZ - kubp[i].fx * sinZ;
      pnt[i].fx = kubp[i].fx * cosZ + kubp[i].fy * sinZ;

      pnt[i].fz = kubp[i].fz * cosY - pnt[i].fx * sinY;
      pnt[i].fx = (pnt[i].fx * cosY + kubp[i].fz * sinY) * z ;

      float temp = pnt[i].fy * cosX + pnt[i].fz * sinX;
      pnt[i].fz = (pnt[i].fz * cosX - pnt[i].fy * sinX) * z;
      pnt[i].fy = temp * z;                       
	}

   static int rot = 512;

   rot += 4;
   rot &= 1023;

   int sinus = sintab[rot+256];
   int cosin = sintab[rot];

   for(int y=-32; y<32; y++)
	{
      byte *temp = Txt + ((y+32)<<6) + 32;
   
      int siny = ((sinus*y) >> 16 )+32;
      int cosy = ((cosin*y) >> 10 )-(32<<6);

      for(int x=-32; x<32; x++)
		{
         unsigned int xT = (( cosin*x ) >> 16 ) + siny;
         unsigned int yT = (( sinus*x ) >> 10 ) - cosy;
         
         if( xT<64 && yT< (64<<6) )
            temp[x] = Texture[ xT+ (yT & 0xffffffc0) ];
		}
	}

}

void project()
{
   for( int i=0; i<8; i++ )
   {
      pnt[i].x = (( pnt[i].fx / (pnt[i].fz + 15)) * 700) + 160;      
      pnt[i].y = (( pnt[i].fy / (pnt[i].fz + 15)) * 600) + 100;      
   }
}


void load( char *file, byte *data, unsigned int length )
{
   FILE *fp = fopen( file, "rb" );

   if( ferror( fp ) )
   {
      printf("Error opening %s\n", file);
		exit( EXIT_FAILURE );
   }

   fread( data, 1, length, fp);

   if( ferror( fp ) )
   {
      printf("Error reading %s\n", file);
		exit( EXIT_FAILURE );
   }

   fclose( fp );
}

void loadpal( char *file )
{
   FILE *fp = fopen( file, "r" );

   if( ferror( fp ) )
   {
      printf("Error opening %s\n", file);
		exit( EXIT_FAILURE );
   }

   for( int i=0; i<256; i++)
   {
      int r,g,b;

      fscanf( fp, "%d %d %d", &r,&g,&b);

      pal[i][0] = r>>2;
      pal[i][1] = g>>2;
      pal[i][2] = b>>2;
   }

   if( ferror( fp ) )
   {
      printf("Error reading %s\n", file);
		exit( EXIT_FAILURE );
   }

   fclose( fp );
}
